import os

from autonomous import log
from autonomous.ai.baseagent import BaseAgent
from autonomous.model.autoattr import StringAttr


class AudioAgent(BaseAgent):
    name = StringAttr(default="audioagent")

    provider = StringAttr(default="local")

    instructions = StringAttr(
        default="You are highly skilled AI trained to assist with generating audio files."
    )
    description = StringAttr(
        default="A helpful AI assistant trained to assist with generating audio files."
    )

    def generate(self, prompt, voice=None):
        return self.get_client(
            os.environ.get("TTS_AI_AGENT", self.provider)
        ).generate_audio(prompt, voice=voice)

    def transcribe(
        self,
        audio,
        prompt="Transcribe this audio clip",
        display_name="audio.opus",
        whisper_context="",
        evaluation=False,
    ):
        # self.add_to_job_meta('prompt', prompt)
        return self.get_client(
            os.environ.get("TRANSCRIBE_AI_AGENT", self.provider)
        ).generate_transcription(audio, prompt=prompt, evaluation=evaluation)

    def available_voices(self, filters=[]):
        return self.get_client().list_voices(filters=filters)



import io
import json
import os
import random

import markdown
import requests
from PIL import Image
from pydub import AudioSegment

from autonomous import log
from autonomous.model.autoattr import ListAttr, StringAttr
from autonomous.model.automodel import AutoModel


class LocalAIModel(AutoModel):
    messages = ListAttr(StringAttr(default=[]))
    name = StringAttr(default="agent")
    instructions = StringAttr(default="You are a helpful AI.")
    description = StringAttr(default="A Local AI Model using Ollama and Media AI.")

    # Config
    _ollama_url = os.environ.get("OLLAMA_API_BASE_URL", "")
    _media_url = os.environ.get("MEDIA_API_BASE_URL", "")
    _audio_url = os.environ.get("MEDIA_AUDIO_API_BASE_URL", "")
    _image_url = os.environ.get("MEDIA_IMAGE_API_BASE_URL", "")
    _text_model = os.environ.get("OLLAMA_TEXT_MODEL", "gemma2:27b")
    _json_model = os.environ.get("OLLAMA_JSON_MODEL", "hermes3")

    VOICES = {
        "Zephyr": ["female"],
        "Puck": ["male"],
        "Charon": ["male"],
        "Kore": ["female"],
        "Fenrir": ["non-binary"],
        "Leda": ["female"],
        "Orus": ["male"],
        "Aoede": ["female"],
        "Callirhoe": ["female"],
        "Autonoe": ["female"],
        "Enceladus": ["male"],
        "Iapetus": ["male"],
        "Umbriel": ["male"],
        "Algieba": ["male"],
        "Despina": ["female"],
        "Erinome": ["female"],
        "Algenib": ["male"],
        "Rasalgethi": ["non-binary"],
        "Laomedeia": ["female"],
        "Achernar": ["female"],
        "Alnilam": ["male"],
        "Schedar": ["male"],
        "Gacrux": ["female"],
        "Pulcherrima": ["non-binary"],
        "Achird": ["male"],
        "Zubenelgenubi": ["male"],
        "Vindemiatrix": ["female"],
        "Sadachbia": ["male"],
        "Sadaltager": ["male"],
        "Sulafar": ["female"],
    }

    def _convert_tools_to_json_schema(self, user_function):
        schema = {
            "name": user_function.get("name"),
            "description": user_function.get("description", ""),
            "parameters": user_function.get("parameters", {}),
        }
        return json.dumps(schema, indent=2)

    def _clean_json_response(self, text):
        """
        Robust cleaner for Llama 3 outputs.
        It handles markdown blocks, chatter before/after, and malformed endings.
        """
        text = text.strip()

        # 1. Strip Markdown Code Blocks (```json ... ```)
        if "```" in text:
            import re

            # Regex to capture content inside ```json ... ``` or just ``` ... ```
            # flags=re.DOTALL allows . to match newlines
            pattern = r"```(?:json)?\s*(\{.*?\})\s*```"
            match = re.search(pattern, text, re.DOTALL)
            if match:
                text = match.group(1)
            else:
                # Fallback: simple finding if regex fails due to weird chars
                start = text.find("```")
                end = text.rfind("```")
                # Adjust start to skip the "json" part if present
                first_newline = text.find("\n", start)
                if first_newline != -1 and first_newline < end:
                    text = text[first_newline:end]

        # 2. Heuristic extraction: Find the first '{' and the last '}'
        # This fixes cases where Llama says "Here is the JSON: { ... }"
        start_idx = text.find("{")
        end_idx = text.rfind("}")

        if start_idx != -1 and end_idx != -1:
            text = text[start_idx : end_idx + 1]

        return text.strip()

    def _evaluate_response(instruction, content, goal):
        eval_prompt = {
            "model": os.environ.get("OLLAMA_TEXT_MODEL", "gemma2:27b"),
            "messages": [
                {
                    "role": "user",
                    "content": f"Evaluate the original instruction and goal in regards to the output produced.\n\nGoal: {goal}\nInstruction: {instruction}\n\nOutput:\n{content}\n\n Write an improved prompt that will get the output closer to the stated goal. Only output the complete improved prompt. Do not preface it with any message.",
                }
            ],
            "stream": False,
        }
        res_eval = requests.post("http://ollama:11434/api/chat", json=eval_prompt)

        return res_eval.json().get("message", {}).get("content", "")

    def generate_json(
        self,
        message,
        system_prompt=None,
        uri="",
        context={},
        evaluation=False,
    ):
        system_prompt = system_prompt or self.instructions
        full_system_prompt = (
            f"{system_prompt}.\n"
            f"You are a strict JSON generator. Output ONLY a valid JSON object matching the given schema.\n"
            f"IMPORTANT RULES:\n"
            f"1. Do not include markdown formatting or explanations.\n"
            f"2. DOUBLE CHECK nested quotes inside strings. Escape them properly.\n"
            f"3. Ensure all arrays and objects are closed.\n"
        )

        # 2. Add Context (Applies to all strategies)
        if context:
            full_system_prompt += (
                f"\n\n### GROUND TRUTH CONTEXT ###\n"
                f"Adhere strictly to this context:\n"
                f"{json.dumps(context, indent=2)}"
            )
        if uri:
            full_system_prompt += f"Use the following URI for reference: {uri}"

        # 3. Payload Construction
        payload = {
            "model": self._json_model,
            "messages": [
                {"role": "system", "content": full_system_prompt},
                {"role": "user", "content": message},
            ],
            "format": "json",
            "stream": False,
            "keep_alive": "24h",
            "options": {
                "num_ctx": 8192,
                "temperature": 0.9,  # Keep high for creativity
                "top_p": 0.9,
                "repeat_penalty": 1.1,
            },
        }

        log("==== LocalAI JSON Payload ====", payload, _print=True)

        result_text = ""
        try:
            response = requests.post(f"{self._ollama_url}/chat", json=payload)
            log(response)
            response.raise_for_status()

            result_text = response.json().get("message", {}).get("content", "{}")

            # Clean & Parse
            clean_text = self._clean_json_response(result_text)
            result_dict = json.loads(clean_text)

            # Unwrap (Handle cases where model wraps in 'parameters' key)
            if "parameters" in result_dict and isinstance(
                result_dict["parameters"], dict
            ):
                params = result_dict.pop("parameters")
                result_dict.update(params)
            if evaluation:
                eval_res = self._evaluate_response(
                    message, response, full_system_prompt
                )
                log(f"Evaluation:\n {eval_res}")
                payload["messages"][1]["content"] = eval_res
                response = requests.post(f"{self._ollama_url}/chat", json=payload)
                response.raise_for_status()
                result_text = response.json().get("message", {}).get("content", "{}")
                # Clean & Parse
                clean_text = self._clean_json_response(result_text)
                result_dict = json.loads(clean_text)

                # Unwrap (Handle cases where model wraps in 'parameters' key)
                if "parameters" in result_dict and isinstance(
                    result_dict["parameters"], dict
                ):
                    params = result_dict.pop("parameters")
                    result_dict.update(params)
            log("==== LocalAI JSON Result ====", result_dict, _print=True)
            return result_dict

        except Exception as e:
            log(f"==== LocalAI JSON Error: {e} ====", _print=True)
            if result_text:
                log(
                    f"--- FAILED RAW OUTPUT ---\n{result_text}\n-----------------------",
                    _print=True,
                )
            return {}

    def generate_text(
        self,
        message,
        additional_instructions="",
        uri="",
        context={},
        temperature=0.9,
        evaluation=False,
    ):
        # 1. Base System Prompt

        if context:
            additional_instructions += (
                f"\n\n### GROUND TRUTH CONTEXT ###\n"
                f"The following context is absolute truth for this interaction. "
                f"Prioritize it over your internal training data. "
                f"If the context says the sky is green, it is green.\n"
                f"{json.dumps(context, indent=2)}"
            )
        if uri:
            additional_instructions += f"\nUse the following URI for reference: {uri}"

        # 3. Send to Ollama
        payload = {
            "model": self._text_model,
            "messages": [
                {"role": "system", "content": additional_instructions},
                {"role": "user", "content": message},
            ],
            "stream": False,
            "keep_alive": "24h",
            "options": {
                "num_ctx": 8192,
                # 1. CREATIVITY
                "temperature": temperature,
                # 2. PREVENT CUTOFFS
                "num_predict": -1,
                # 3. PREVENT LOOPS
                "repeat_penalty": 1.1,
            },
        }

        try:
            response = requests.post(f"{self._ollama_url}/chat", json=payload)
            response.raise_for_status()
            if evaluation:
                eval_res = self._evaluate_response(
                    message, response, additional_instructions
                )
                log(f"Evaluation:\n {eval_res}")
                payload["messages"][1]["content"] = eval_res
                response = requests.post(f"{self._ollama_url}/chat", json=payload)
                response.raise_for_status()
            return response.json().get("message", {}).get("content", "")
        except Exception as e:
            log(f"==== LocalAI Text Error: {e} ====", _print=True)
            return "Error generating text."

    def summarize_text(self, text, primer=""):
        primer = primer or "Summarize the following text concisely."
        max_chars = 12000
        chunks = [text[i : i + max_chars] for i in range(0, len(text), max_chars)]

        full_summary = ""
        for chunk in chunks:
            payload = {
                "model": self._text_model,
                "messages": [{"role": "user", "content": f"{primer}:\n\n{chunk}"}],
                "stream": False,
                "keep_alive": "24h",
            }
            try:
                log(f"Payload sent: {payload}...", _print=True)
                res = requests.post(f"{self._ollama_url}/chat", json=payload)
                full_summary += res.json().get("message", {}).get("content", "") + "\n"
                log(f"Chunk summarized: {full_summary}.", _print=True)
            except Exception as e:
                log(f"Summary Error: {e}", _print=True)
                break

        return full_summary

    def generate_transcription(
        self, audio_file, prompt="", whisper_context="", evaluation=False
    ):
        if isinstance(audio_file, bytes):
            f_obj = io.BytesIO(audio_file)
        else:
            f_obj = audio_file

        files = {"file": ("audio.opus", f_obj, "audio/ogg")}

        # Whisper uses this to "continue" the sentence style.
        data = {"prompt": whisper_context} if whisper_context else {}

        response = requests.post(
            f"{self._audio_url}/transcribe",
            files=files,
            data=data,
        )
        response.raise_for_status()

        response_text = response.json().get("text", "")
        full_llm_prompt = f"{prompt}\n\nPREVIOUS CONTEXT: {whisper_context}\n\nRAW TRANSCRIPT:\n{response_text}"
        log(
            "==== TRANSCRIPTION PROMPT ====",
            full_llm_prompt,
            f"TOKENS: {len(full_llm_prompt.split())}",
            _print=True,
        )
        system_prompt = """
You are an expert Scribe and Editor. Your task is to transform a raw, automated audio transcript into a clean, readable script format.

GUIDELINES:
**Speaker Identification**: Use the Context to guess who is speaking to the best of your ability.
**Cleanup**: Remove verbal tics (um, uh, like, you know) and stuttering. Fix punctuation. Remove tangent conversations not relevant to the topic, such as "What did you do last weekend?", "Hand me some chips", etc.
**NO PREAMBLE**: Output ONLY the Markdown formatted script. Do not add introductory or concluding remarks.
"""
        try:
            result = self.generate_text(
                full_llm_prompt, additional_instructions=system_prompt, temperature=0.1
            )
            if evaluation:
                eval_res = self._evaluate_response(
                    full_llm_prompt, result, system_prompt
                )
                log(f"Evaluation:\n {eval_res}")
                result = self.generate_text(
                    eval_res,
                    additional_instructions=system_prompt,
                    temperature=0.1,
                )
            log(f"FINAL TRANSCRIPTION CHUNK: {result}", _print=True)
            return markdown.markdown(result)
        except Exception as e:
            log(f"STT Error: {e}", _print=True)
            raise e

    def list_voices(self, filters=[]):
        if not filters:
            return list(self.VOICES.keys())
        voices = []
        for voice, attribs in self.VOICES.items():
            if any(f.lower() in attribs for f in filters):
                voices.append(voice)
        return voices

    def generate_audio(
        self,
        prompt,
        voice=None,
    ):
        voice = voice or random.choice(list(self.VOICES.keys()))
        try:
            payload = {"text": prompt, "voice": voice}
            response = requests.post(f"{self._media_url}/tts", json=payload)
            response.raise_for_status()
            wav_bytes = response.content
            audio = AudioSegment.from_file(io.BytesIO(wav_bytes), format="wav")
            audio_buffer = io.BytesIO()
            audio.export(audio_buffer, format="opus")
            return audio_buffer.getvalue()
        except Exception as e:
            log(f"TTS Error: {e}", _print=True)
            return None

    def _get_dimensions(self, aspect_ratio):
        """
        Returns a tuple: ((base_w, base_h), (target_w, target_h))

        1. base_*: The resolution sent to SDXL (approx 1024x1024).
        2. target_*: The final resolution to resize/upscale to.
        """
        # Standard SDXL buckets (approx 1MP)
        # We use these for the initial generation to ensure good composition.
        sdxl_base = {
            "1:1": (1024, 1024),
            "Portrait": (896, 1152),  # 3:4
            "Landscape": (1216, 832),  # 3:2 or 16:9 approx
        }

        # The Logic: Define the target, map it to the closest SDXL base
        # Format: "Key": ((Base_W, Base_H), (Target_W, Target_H))
        resolutions = {
            # Standard
            "1:1": ((832, 832), (1024, 1024)),
            "3:4": ((832, 1152), (1664, 2304)),
            "4:3": ((1152, 832), (2304, 1664)),
            # High Res (The logic changes here)
            "16:9": ((1216, 832), (2048, 1152)),
            "9:16": ((832, 1216), (1152, 2048)),
            # 2K Tier
            "2K": ((1216, 832), (2048, 1152)),  # Base is 1216x832 -> Upscale to 2K
            "2KPortrait": ((832, 1216), (1152, 2048)),
            # 4K Tier (The generated image will be upscaled ~3x)
            "4K": ((1216, 832), (3840, 2160)),
            "4KPortrait": ((832, 1216), (2160, 3840)),
        }

        # Default to 1:1 if unknown
        return resolutions.get(aspect_ratio, ((832, 832), (1024, 1024)))

    def generate_image(
        self, prompt, negative_prompt="", files=None, aspect_ratio="2KPortrait"
    ):
        # 1. Resolution Calculation
        (base_w, base_h), (target_w, target_h) = self._get_dimensions(aspect_ratio)

        # 2. Construct Base Generation Payload
        # We tell the AI to generate the smaller, stable size first.
        data = {
            "prompt": prompt,
            "negative_prompt": negative_prompt,
            "aspect_ratio": aspect_ratio,
            "width": base_w,
            "height": base_h,
        }

        try:
            # Handle Input Files (for Img2Img)
            files_list = []
            if files and isinstance(files, dict):
                for fn, f_bytes in files.items():
                    if isinstance(f_bytes, bytes):
                        file_obj = io.BytesIO(f_bytes)
                    else:
                        file_obj = f_bytes
                    files_list.append(("files", (fn, file_obj, "image/png")))

            # 3. Step 1: Generate Base Image
            url = f"{self._image_url}/generate-image"
            if files_list:
                response = requests.post(url, data=data, files=files_list)
            else:
                response = requests.post(url, data=data)

            response.raise_for_status()
            image_content = response.content

            # 4. Step 2: Upscale (If necessary)
            if (base_w, base_h) != (target_w, target_h):
                log(
                    f"Requesting AI Upscale: {base_w}x{base_h} -> {target_w}x{target_h}...",
                    _print=True,
                )

                # Prepare payload for the /upscale route
                upscale_data = {
                    "prompt": prompt,  # Reuse prompt to guide texture generation
                    "width": target_w,  # Explicitly tell server the target size
                    "height": target_h,
                }

                # Send the image we just generated back to the server as a file
                upscale_files = {
                    "file": ("generated.png", io.BytesIO(image_content), "image/png")
                }

                upscale_response = requests.post(
                    f"{self._image_url}/upscale", data=upscale_data, files=upscale_files
                )
                upscale_response.raise_for_status()
                image_content = upscale_response.content

            log("==== LocalAI Image Generation Complete ====", data, _print=True)
            return image_content

        except Exception as e:
            log(f"Image Gen Error: {e}", _print=True)
            return None



from autonomous import log
from autonomous.model.autoattr import ReferenceAttr, StringAttr
from autonomous.model.automodel import AutoModel
from autonomous.taskrunner.autotasks import AutoTasks

from .models.gemini import GeminiAIModel
from .models.local_model import LocalAIModel


class BaseAgent(AutoModel):
    meta = {"abstract": True, "allow_inheritance": True, "strict": False}

    # 2. Map string names to classes
    MODEL_REGISTRY = {
        "local": LocalAIModel,
        "gemini": GeminiAIModel,
    }

    # 3. Allow client to be ANY of the supported models
    client = ReferenceAttr(choices=[LocalAIModel, GeminiAIModel])

    # 4. Add a provider field (default to local, can be overridden per agent)
    provider = StringAttr(default="gemini")

    def delete(self):
        if self.client:
            self.client.delete()
        return super().delete()

    def get_agent_id(self):
        return self.get_client().id

    def get_client(self, provider=None):
        # 5. Determine which class to use based on the provider string
        model_class = self.MODEL_REGISTRY.get(provider or self.provider, LocalAIModel)
        # If we already have a client, but it's the WRONG type (e.g. we switched providers), we might want to re-instantiate. For simplicity, we check if it exists first.
        if not isinstance(self.client, model_class):
            if self.client:
                log(
                    f"Re-instantiating client for agent {self.name} from {type(self.client).__name__} to {model_class.__name__}"
                )
                self.client.delete()
            self.client = model_class(
                name=self.name,
                instructions=self.instructions,
                description=self.description,
            )
            self.client.save()
            self.save()

        return self.client

    # def add_to_job_meta(self, k, v):
    #     if job := AutoTasks.get_current_task():
    #         job.meta(k, v)



import os
import time
import urllib.parse
import uuid
from datetime import datetime

import numpy as np
import pymongo
import redis
import requests

from autonomous import log
from autonomous.taskrunner.autotasks import AutoTasks, TaskPriority

# CONFIGURATION
db_host = os.getenv("DB_HOST", "db")
db_port = os.getenv("DB_PORT", 27017)
password = urllib.parse.quote_plus(str(os.getenv("DB_PASSWORD")))
username = urllib.parse.quote_plus(str(os.getenv("DB_USERNAME")))
MONGO_URI = f"mongodb://{username}:{password}@{db_host}:{db_port}/?authSource=admin"
MEDIA_URL = os.getenv("MEDIA_API_BASE_URL", "http://media_ai:5005")
REDIS_HOST = os.getenv("REDIS_HOST", "cachedb")
REDIS_PORT = int(os.getenv("REDIS_PORT", 6379))
# DB SETUP
r = redis.Redis(host=REDIS_HOST, port=REDIS_PORT, decode_responses=True)
mongo = pymongo.MongoClient(MONGO_URI)
db = mongo[os.getenv("DB_DB")]
# connect(host=f"mongodb://{username}:{password}@{host}:{port}/{dbname}?authSource=admin")


def get_vector(text):
    """Helper to get embedding from your Media AI container"""
    try:
        resp = requests.post(f"{MEDIA_URL}/embeddings", json={"text": text}, timeout=30)
        if resp.status_code == 200:
            return resp.json()["embedding"]
    except Exception as e:
        print(f"Vector Gen Failed: {e}")
    return None


def process_single_object_sync(object_id, collection_name, token):
    """
    THE WORKER FUNCTION (Runs in Background).
    It is safe to sleep here because we are not in the web request.
    """
    str_id = str(object_id)
    token_key = f"sync_token:{collection_name}:{str_id}"

    # 1. THE DEBOUNCE WAIT (Happens in background)
    time.sleep(5)

    # 2. THE VERIFICATION
    # Check if a newer save happened while we slept
    current_active_token = r.get(token_key)

    if current_active_token != token:
        print(f"Skipping sync for {str_id}: Superseded by a newer save.")
        return

    # 3. THE EXECUTION (Embedding generation)
    # print(f"Processing Sync for: {str_id} in {collection_name}")

    from bson.objectid import ObjectId

    # FIX: Use dynamic collection access instead of db.objects
    try:
        # Tries to convert string ID to ObjectId.
        # If your DB uses String IDs, remove the ObjectId() wrapper.
        oid = ObjectId(object_id)
        doc = db[collection_name].find_one({"_id": oid})
    except Exception:
        # Fallback if ID is not a valid ObjectId string
        doc = db[collection_name].find_one({"_id": object_id})

    if not doc:
        print(f"Object {object_id} not found in collection '{collection_name}'")
        # Optional: Remove from Redis index if it exists
        r.delete(f"{collection_name}:{object_id}")
        return

    # 2. Construct Searchable Text
    # (Existing logic...)
    log(doc.get("associations"))
    searchable_text = (
        f"{doc.get('name', '')}: {doc.get('history', '')}"
        f"Related: {', '.join([str(a) for a in doc['associations'][:20]]) if doc.get('associations') else ''}"
    )

    if len(searchable_text) < 10:
        return

    # 3. Generate Vector
    vector = get_vector(searchable_text)

    # 4. Save to Redis Index
    if vector:
        r.hset(
            f"lore:{object_id}",
            mapping={
                "mongo_id": str(object_id),
                "collection": collection_name,  # Useful for debugging
                "content": searchable_text,
                "vector": np.array(vector, dtype=np.float32).tobytes(),
                "last_synced": datetime.utcnow().isoformat(),
            },
        )
        print(f"Successfully Indexed: {doc.get('name')}")


def request_indexing(object_id, collection_name):
    """
    THE TRIGGER FUNCTION (Runs in Main App).
    MUST BE FAST. NO SLEEPING HERE.
    """
    print("Requesting Indexing...")

    # Initialize the Task Runner
    task_runner = AutoTasks()

    str_id = str(object_id)
    token_key = f"sync_token:{collection_name}:{str_id}"

    # 1. GENERATE NEW TOKEN
    current_token = str(uuid.uuid4())

    # 2. SAVE TOKEN TO REDIS (Instant)
    r.set(token_key, current_token, ex=300)

    # 3. ENQUEUE THE TASK (Instant)
    try:
        task_runner.task(
            process_single_object_sync,  # The function to run later
            object_id=str_id,
            collection_name=collection_name,
            token=current_token,
            priority=TaskPriority.LOW,
        )
        return True
    except Exception as e:
        print(f"Sync Enqueue failed: {e}")
        return False



import os
from enum import Enum

from redis import Redis
from rq import Queue, get_current_job
from rq.command import send_stop_job_command
from rq.job import Job


# 1. Define Priorities clearly
class TaskPriority(Enum):
    HIGH = "high"
    DEFAULT = "default"
    LOW = "low"
    IMAGE = "image"
    AUDIO = "audio"


class AutoTask:
    def __init__(self, job):
        self.job = job

    @property
    def id(self):
        return self.job.id

    @property
    def args(self):
        return self.job.args

    @property
    def kwargs(self):
        return self.job.kwargs

    @property
    def status(self):
        return self.job.get_status()

    @property
    def func_name(self):
        return self.job.func_name

    @property
    def timeout(self):
        return self.job.timeout

    @property
    def enqueued_at(self):
        return self.job.enqueued_at

    @property
    def started_at(self):
        return self.job.started_at

    @property
    def ended_at(self):
        return self.job.ended_at

    @property
    def retries_left(self):
        return self.job.retries_left

    @property
    def is_failed(self):
        return self.job.get_status() == "failed"

    @property
    def is_finished(self):
        return self.job.get_status() == "finished"

    @property
    def origin(self):
        return self.job.origin

    @property
    def exc_info(self):
        return self.job.exc_info

    @property
    def result(self):
        return self.job.result

    def delete(self):
        self.job.delete()

    def meta(self, key=None, value=None):
        if value:
            self.job.meta[key] = value
            self.job.save_meta()
        if key:
            return self.job.meta.get(key, "")
        return self.job.meta


class AutoTasks:
    _connection = None
    # We remove the single 'queue' class attribute because we now have multiple

    config = {
        "host": os.environ.get("REDIS_HOST", "cachedb"),
        "port": os.environ.get("REDIS_PORT", 6379),
        "password": os.environ.get("REDIS_PASSWORD"),
        "username": os.environ.get("REDIS_USERNAME"),
        "db": os.environ.get("REDIS_DB", 0),
    }

    def __init__(self):
        # Establish connection once (Singleton pattern logic)
        if not AutoTasks._connection:
            options = {}
            if AutoTasks.config.get("password"):
                options["password"] = AutoTasks.config.get("password")

            AutoTasks._connection = Redis(
                host=AutoTasks.config.get("host"),
                port=AutoTasks.config.get("port"),
                decode_responses=False,
                **options,
            )

    def _get_queue(self, priority_name):
        """Helper to get or create the queue object for a specific priority"""
        return Queue(priority_name, connection=AutoTasks._connection)

    def task(self, func, *args, **kwargs):
        """
        Enqueues a job.
        kwarg 'priority' determines the queue (default: 'default').
        """
        job_timeout = kwargs.pop("_task_job_timeout", 7200)

        # 2. Extract Priority (support Enum or string)
        priority = kwargs.pop("priority", TaskPriority.DEFAULT)
        queue_name = priority.value if isinstance(priority, TaskPriority) else priority

        # 3. Get the specific queue
        q = self._get_queue(queue_name)

        # 4. Enqueue
        job = q.enqueue(func, args=args, kwargs=kwargs, job_timeout=job_timeout)

        return AutoTask(job)

    def get_current_task(self):
        if job := get_current_job():
            return AutoTask(job)

    def get_task(self, job_id):
        try:
            if job := Job.fetch(job_id, connection=AutoTasks._connection):
                return AutoTask(job)
        except Exception:
            return None

    def get_tasks(self):

        high_queue = Queue("high", connection=self._connection)
        default_queue = Queue("default", connection=self._connection)
        low_queue = Queue("low", connection=self._connection)
        audio_queue = Queue("audio", connection=self._connection)
        image_queue = Queue("image", connection=self._connection)

        registries = {
            "started": [
                high_queue.started_job_registry,
                default_queue.started_job_registry,
                low_queue.started_job_registry,
                audio_queue.started_job_registry,
                image_queue.started_job_registry,
            ],
            "finished": [
                high_queue.finished_job_registry,
                default_queue.finished_job_registry,
                low_queue.finished_job_registry,
                audio_queue.finished_job_registry,
                image_queue.finished_job_registry,
            ],
            "failed": [
                high_queue.failed_job_registry,
                default_queue.failed_job_registry,
                low_queue.failed_job_registry,
                audio_queue.failed_job_registry,
                image_queue.failed_job_registry,
            ],
            "queued": [high_queue, default_queue, low_queue, audio_queue, image_queue],
        }

        tasks = {}
        for status, regs in registries.items():
            all_job_ids = []
            for reg in regs:
                all_job_ids.extend(reg.get_job_ids())
            # Use a set to remove duplicate job_ids if a job is in multiple registries
            unique_job_ids = sorted(list(set(all_job_ids)), reverse=True)
            jobs = Job.fetch_many(unique_job_ids, connection=self._connection)
            # Filter out None values in case a job expired between fetch and get
            tasks[status] = [AutoTask(job) for job in jobs if job]
        return tasks

    def kill(self, pk):
        job = self.get_task(pk)
        if job.status == "started":
            send_stop_job_command(AutoTasks._connection, job.id)
        job.delete()



import base64
import io
import json

from autonomous import log
from autonomous.model.autoattr import ListAttr, StringAttr
from autonomous.model.automodel import AutoModel


class MockAIModel(AutoModel):
    """
    A Offline Mock Model for development.
    Returns valid, structure-compliant dummy data instantly.
    """

    messages = ListAttr(StringAttr(default=[]))
    name = StringAttr(default="mock_agent")
    instructions = StringAttr(default="You are a Mock AI.")
    description = StringAttr(default="Offline Mock Provider for testing.")

    # A 1x1 Red Pixel PNG (Base64 decoded to bytes)
    # This ensures PIL/Frontend code receives valid image data.
    _MOCK_IMAGE_BYTES = base64.b64decode(
        "iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAYAAAAfFcSJAAAADUlEQVR42mP8z8BQDwAEhQGAhKmMIQAAAABJRU5ErkJggg=="
    )

    # A 1-second silent WAV file (Base64 decoded to bytes)
    _MOCK_AUDIO_BYTES = base64.b64decode(
        "UklGRigAAABXQVZFZm10IBAAAAABAAEAQB8AAEAfAAABAAgAZGF0YQAAAAEA"
    )

    def generate_json(self, message, system_prompt=None, uri="", context={}):
        """
        Returns a rich dummy JSON object.
        It attempts to guess the context based on keys found in the prompt,
        otherwise returns a safe default TTRPG object.
        """
        log(f"⚡ [MOCK] Generating JSON for prompt: {message[:50]}...", _print=True)

        # 1. Default Mock Object
        mock_response = {}

        # 2. Heuristic: If the user provided a context with 'name', use it.
        # This makes the mock feel slightly responsive.
        if context:
            if "name" in context:
                mock_response["name"] = f"Mock {context['name']}"
            if "title" in context:
                mock_response["name"] = f"Mock {context['title']}"

        return mock_response

    def generate_text(self, message, additional_instructions="", uri="", context={}):
        log(f"⚡ [MOCK] Generating Text...", _print=True)
        return (
            "Lorem ipsum dolor sit amet, consectetur adipiscing elit. "
            "This is MOCK TEXT generated because you are currently offline. "
            "The quick brown fox jumps over the lazy dog."
        )

    def summarize_text(self, text, primer=""):
        log(f"⚡ [MOCK] Summarizing {len(text)} chars...", _print=True)
        return "This is a mock summary of the provided text. It is concise and offline."

    def generate_transcription(self, audio_file, prompt=""):
        log(f"⚡ [MOCK] Transcribing Audio...", _print=True)
        return "This is a mock transcription of the audio file."

    def generate_audio(self, prompt, voice=None):
        log(f"⚡ [MOCK] Generating Audio...", _print=True)
        return self._MOCK_AUDIO_BYTES

    def generate_image(
        self, prompt, negative_prompt="", files=None, aspect_ratio="2KPortrait"
    ):
        log(f"⚡ [MOCK] Generating Image ({aspect_ratio})...", _print=True)
        # Returns a valid 1x1 pixel PNG so your UI displays something red instead of crashing
        return self._MOCK_IMAGE_BYTES



import os
import urllib.parse
from datetime import datetime

import bson
from autonomous.db import Document, connect, db_sync, signals
from autonomous.db.errors import ValidationError
from autonomous.db.fields import DateTimeField

from autonomous import log

host = os.getenv("DB_HOST", "db")
port = os.getenv("DB_PORT", 27017)
password = urllib.parse.quote_plus(str(os.getenv("DB_PASSWORD")))
username = urllib.parse.quote_plus(str(os.getenv("DB_USERNAME")))
dbname = os.getenv("DB_DB")
# log(f"Connecting to MongoDB at {host}:{port} with {username}:{password} for {dbname}")
connect(host=f"mongodb://{username}:{password}@{host}:{port}/{dbname}?authSource=admin")


class AutoModel(Document):
    meta = {"abstract": True, "allow_inheritance": True, "strict": False}
    last_updated = DateTimeField(default=datetime.now)

    def __eq__(self, other):
        return str(self.pk) == str(other.pk) if other else False

    def __lt__(self, other):
        return str(self.pk) < str(other.pk) if other else False

    def __gt__(self, other):
        return not (str(self.pk) < str(other.pk)) if other else False

    def __le__(self, other):
        return self < other or self == other

    def __ge__(self, other):
        return self > other or self == other

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(
            (
                self.model_name(),
                self.pk,
            )
        )

    @classmethod
    def auto_pre_init(cls, sender, document, **kwargs):
        values = kwargs.pop("values", None)
        if pk := values.get("pk") or values.get("id"):
            # Try to load the existing document from the database
            if existing_doc := sender._get_collection().find_one(
                {"_id": bson.ObjectId(pk)}
            ):
                # Update the current instance with the existing data
                existing_doc.pop("_id", None)
                existing_doc.pop("_cls", None)
                for k, v in existing_doc.items():
                    if not values.get(k):
                        values[k] = v

    @classmethod
    def _auto_pre_init(cls, sender, document, **kwargs):
        sender.auto_pre_init(sender, document, **kwargs)

    @classmethod
    def auto_post_init(cls, sender, document, **kwargs):
        document.last_updated = datetime.now()

    @classmethod
    def _auto_post_init(cls, sender, document, **kwargs):
        sender.auto_post_init(sender, document, **kwargs)

    def model_name(self, qualified=False):
        """
        Get the fully qualified name of this model.

        Returns:
            str: The fully qualified name of this model.
        """
        return (
            f"{self.__module__}.{self._class_name}" if qualified else self._class_name
        )

    @classmethod
    def get_model(cls, model, pk=None):
        try:
            Model = cls.load_model(model)
        except ValueError:
            Model = None
        return Model.get(pk) if Model and pk else Model

    @classmethod
    def load_model(cls, model):
        if not isinstance(model, str):
            return model

        subclasses = AutoModel.__subclasses__()
        visited_subclasses = []
        while subclasses:
            # log(subclasses, _print=True)
            subclass = subclasses.pop()
            if "_meta" in subclass.__dict__ and not subclass._meta.get("abstract"):
                # log(f"Checking {subclass.__name__}", _print=True)
                if subclass.__name__.lower() == model.lower():
                    return subclass
            if subclass not in visited_subclasses:
                subclasses += subclass.__subclasses__()
                visited_subclasses += [subclass]
        raise ValueError(f"Model {model} not found")

    @classmethod
    def get(cls, pk):
        """
        Get a model by primary key.

        Args:
            pk (int): The primary key of the model to retrieve.

        Returns:
            AutoModel or None: The retrieved AutoModel instance, or None if not found.
        """

        if isinstance(pk, str):
            try:
                pk = bson.ObjectId(pk)
            except bson.errors.InvalidId:
                return None
        elif isinstance(pk, dict) and "$oid" in pk:
            pk = bson.ObjectId(pk["$oid"])
        try:
            # log(pk, type(pk))
            result = cls.objects.get(id=pk)
            # log(result)
        except cls.DoesNotExist as e:
            # log(f"Model {cls.__name__} with pk {pk} not found : {e}")
            return None
        except ValidationError as e:
            # traceback.print_stack(limit=5)
            log(f"Model Validation failure {cls.__name__} [{pk}]: {e}")
            return None
        except Exception as e:
            log(f"Error getting model {cls.__name__} with pk {pk}: {e}", _print=True)
            raise e
        else:
            return result

    @classmethod
    def random(cls):
        """
        Get a model by primary key.

        Args:
            pk (int): The primary key of the model to retrieve.

        Returns:
            AutoModel or None: The retrieved AutoModel instance, or None if not found.
        """
        pipeline = [{"$sample": {"size": 1}}]

        result = cls.objects.aggregate(pipeline)
        random_document = next(result, None)
        return cls._from_son(random_document) if random_document else None

    @classmethod
    def all(cls):
        """
        Get all models of this type.

        Returns:
            list: A list of AutoModel instances.
        """
        return list(cls.objects())

    @classmethod
    def search(cls, _order_by=None, _limit=None, **kwargs):
        """
        Search for models containing the keyword values.

        Args:
            **kwargs: Keyword arguments to search for (dict).

        Returns:
            list: A list of AutoModel instances that match the search criteria.
        """
        new_kwargs = {}
        for k, v in kwargs.items():
            if isinstance(v, str):
                new_k = f"{k}__icontains"
                new_kwargs[new_k] = v
            else:
                new_kwargs[k] = v
        results = cls.objects(**new_kwargs)
        if _order_by:
            results = results.order_by(*_order_by)
        if _limit:
            if isinstance(_limit, list):
                results = results[_limit[0] : _limit[-1]]
            else:
                results = results[:_limit]
        return list(results)

    @classmethod
    def find(cls, **kwargs):
        """
        Find the first model containing the keyword values and return it.

        Args:
            **kwargs: Keyword arguments to search for (dict).

        Returns:
            AutoModel or None: The first matching AutoModel instance, or None if not found.
        """
        return cls.objects(**kwargs).first()

    @classmethod
    def auto_pre_save(cls, sender, document, **kwargs):
        """
        Post-save hook for this model.
        """
        pass

    @classmethod
    def _auto_pre_save(cls, sender, document, **kwargs):
        """
        Post-save hook for this model.
        """
        sender.auto_pre_save(sender, document, **kwargs)

    def save(self, sync=False):
        """
        Save this model to the database.

        Returns:
            int: The primary key (pk) of the saved model.
        """
        obj = super().save()

        if sync:
            db_sync.request_indexing(
                self.pk, collection_name=self._get_collection_name()
            )

        return self.pk

    @classmethod
    def auto_post_save(cls, sender, document, **kwargs):
        """
        Post-save hook for this model.
        """
        pass

    @classmethod
    def _auto_post_save(cls, sender, document, **kwargs):
        """
        Post-save hook for this model.
        """
        sender.auto_post_save(sender, document, **kwargs)

    def delete(self):
        """
        Delete this model from the database.
        """
        return super().delete()


signals.pre_init.connect(AutoModel._auto_pre_init)
signals.post_init.connect(AutoModel._auto_post_init)
signals.pre_save.connect(AutoModel._auto_pre_save)
signals.post_save.connect(AutoModel._auto_post_save)
